// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

// Include this file to access Graph representation implemented using an
// Adjacency List

#include <stdio.h>
#include <ctype.h>

// Used by the mark array
#define UNVISITED 0
#define VISITED 1

#include "link.h"
#include "llist.h"

#include "graph.h"


// Edge implementation for adjacency list
class Edge {
public:
  int vertex, weight;
  Edge() { vertex = -1; weight = -1; }
  Edge(int v, int w) { vertex = v; weight = w; }
};

// Overload for the Edge << operator
ostream& operator << (ostream& s, Edge e)
  { return s << "(" << e.vertex << ", " << e.weight << ")"; }

class Graphl : public Graph {
private:
  int numVertex, numEdge;     // Number of vertices, edges
  List<Edge>** vertex;        // List headers
  int *mark;                  // Pointer to mark array
public:
  Graphl(int numVert) { // Graph has "numVert" vertices
    int i;
    numVertex = numVert;
    numEdge = 0;
    mark = new int[numVert];  // Initialize mark array
    for (i=0; i<numVertex; i++) mark[i] = UNVISITED;
    // Create and initialize adjacency lists
    vertex = (List<Edge>**) new List<Edge>*[numVertex];
    for (i=0; i<numVertex; i++)
      vertex[i] = new LList<Edge>();
  }

  ~Graphl() {       // Destructor
    delete [] mark; // Return dynamically allocated memory
    for (int i=0; i<numVertex; i++) delete [] vertex[i];
    delete [] vertex;
  }

  int n() { return numVertex; } // Number of vertices
  int e() { return numEdge; }   // Number of edges

  // Set edge (v1, v2) to "wgt"
  void setEdge(int v1, int v2, int wgt) {
    Assert(wgt>0, "Illegal weight value");
    Edge it(v2, wgt);
    Edge curr;
    if (vertex[v1]->rightLength() > 0)
      curr = vertex[v1]->getValue();
    else curr.vertex = numVertex;
    // If not already at "v2", we must search from the start
    if (curr.vertex != v2)
      for (vertex[v1]->movetoStart();
           vertex[v1]->rightLength() > 0;
           vertex[v1]->next()) {
        curr = vertex[v1]->getValue(); 
        if (curr.vertex >= v2) break;
      }
    if (curr.vertex == v2)   // Edge (v1, v2) already exists
      curr = vertex[v1]->remove(); // Clear out existing one
    else numEdge++;             // Otherwise, add a new edge
    vertex[v1]->insert(it);
  }
  int getMark(int v) { return mark[v]; }
  void setMark(int v, int val) { mark[v] = val; }

  void delEdge(int v1, int v2) {  // Delete edge (v1, v2)
    Edge curr;
    curr = vertex[v1]->getValue();
    if (curr.vertex != v2)        // If not already at "v2",
      for (vertex[v1]->movetoStart();    // search for start
           vertex[v1]->rightLength() > 0;
           vertex[v1]->next()) {
	curr = vertex[v1]->getValue(); 
        if (curr.vertex >= v2) break;
      }
    // If not at v2 now, then there is no edge (v1, v2)
    if (curr.vertex == v2) {
      curr = vertex[v1]->remove();
      numEdge--;
    }
  }

  int weight(int v1, int v2) { // Return weight of (v1, v2)
    Edge curr;
    if (vertex[v1]->rightLength() > 0)
      curr = vertex[v1]->getValue();
    else curr.vertex = numVertex;
    if (curr.vertex != v2)        // If not already at "v2",
      for (vertex[v1]->movetoStart();    // search for start
           vertex[v1]->rightLength() > 0;
           vertex[v1]->next()) {
	curr = vertex[v1]->getValue(); 
        if (curr.vertex >= v2) break;
      }
    if (curr.vertex == v2)
      return curr.weight;
    else   // If not at "v2" now, there is no edge (v1, v2)
      return 0; // Weight is 0 if (v1, v2) does not exist
  }

  int first(int v) { // Return first neighbor of "v"
    vertex[v]->movetoStart();
    if (vertex[v]->rightLength() == 0)
      return numVertex;      // No neighbor
    Edge it = vertex[v]->getValue();
    return it.vertex;
  }
  // Get v1's next neighbor after "v2"
  int next(int v1, int v2) {
    Edge it;
    // If we're at "v2" now we can just move forward
    if (vertex[v1]->rightLength() > 0) {
      it = vertex[v1]->getValue();
      if (it.vertex == v2) {
        vertex[v1]->next();
        if (vertex[v1]->rightLength() > 0) {
          it = vertex[v1]->getValue();
          return it.vertex;
        }
      }
     }
     // If here, then we need to search for "v2".
     vertex[v1]->movetoStart();
     while (vertex[v1]->rightLength() > 0) {
       it = vertex[v1]->getValue();
       if (it.vertex > v2) break;
       vertex[v1]->next();
     }
    if (vertex[v1]->rightLength() == 0)
      return numVertex; // No vertex
    else return it.vertex;
  }
};

#include "graphutil.cpp"
